"""Module containing the base model"""
import typing
from typing import Dict, Tuple, Any, Type, Union, List, Optional

from pydantic import BaseModel

from pydantic_redis.shared.utils import (
    typing_get_origin,
    typing_get_args,
    from_any_to_str_or_bytes,
    from_dict_to_key_value_list,
    from_bytes_to_str,
    from_str_or_bytes_to_any,
)


from ..store import AbstractStore


class AbstractModel(BaseModel):
    """
    An abstract class to help with typings for Model class
    """

    _primary_key_field: str
    _field_types: Dict[str, Any] = {}
    _store: AbstractStore
    _nested_model_tuple_fields: Dict[str, Tuple[Any, ...]] = {}
    _nested_model_list_fields: Dict[str, Type["AbstractModel"]] = {}
    _nested_model_fields: Dict[str, Type["AbstractModel"]] = {}

    class Config:
        arbitrary_types_allowed = True

    @classmethod
    def get_store(cls) -> AbstractStore:
        """Returns the instance of the store for this model"""
        return cls._store

    @classmethod
    def get_nested_model_tuple_fields(cls):
        """Returns the fields that have tuples of nested models"""
        return cls._nested_model_tuple_fields

    @classmethod
    def get_nested_model_list_fields(cls):
        """Returns the fields that have list of nested models"""
        return cls._nested_model_list_fields

    @classmethod
    def get_nested_model_fields(cls):
        """Returns the fields that have nested models"""
        return cls._nested_model_fields

    @classmethod
    def get_primary_key_field(cls):
        """Gets the protected _primary_key_field"""
        return cls._primary_key_field

    @classmethod
    def get_field_types(cls) -> Dict[str, Any]:
        """Returns the fields types of this model"""
        return cls._field_types

    @classmethod
    def initialize(cls):
        """Initializes class-wide variables for performance's reasons e.g. it caches the nested model fields"""
        cls._field_types = typing.get_type_hints(cls)

        cls._nested_model_list_fields = {}
        cls._nested_model_tuple_fields = {}
        cls._nested_model_fields = {}

        for field, field_type in cls._field_types.items():
            try:
                # In case the annotation is Optional, an alias of Union[X, None], extract the X
                is_generic = hasattr(field_type, "__origin__")
                if (
                    is_generic
                    and typing_get_origin(field_type) == Union
                    and typing_get_args(field_type)[-1] == None.__class__
                ):
                    field_type = typing_get_args(field_type)[0]
                    is_generic = hasattr(field_type, "__origin__")

                if (
                    is_generic
                    and typing_get_origin(field_type) in (List, list)
                    and issubclass(typing_get_args(field_type)[0], AbstractModel)
                ):
                    cls._nested_model_list_fields[field] = typing_get_args(field_type)[
                        0
                    ]

                elif (
                    is_generic
                    and typing_get_origin(field_type) in (Tuple, tuple)
                    and any(
                        [
                            issubclass(v, AbstractModel)
                            for v in typing_get_args(field_type)
                        ]
                    )
                ):
                    cls._nested_model_tuple_fields[field] = typing_get_args(field_type)

                elif issubclass(field_type, AbstractModel):
                    cls._nested_model_fields[field] = field_type

            except (TypeError, AttributeError):
                pass

    @classmethod
    def serialize_partially(cls, data: Optional[Dict[str, Any]]) -> Dict[str, Any]:
        """Converts non primitive data types into str"""
        return {key: from_any_to_str_or_bytes(value) for key, value in data.items()}

    @classmethod
    def deserialize_partially(
        cls, data: Union[List[Any], Dict[Any, Any]] = ()
    ) -> Dict[str, Any]:
        """
        Converts str or bytes to their expected data types, given a list of properties got from the
        list of lists got after calling EVAL on redis.

        EVAL returns a List of Lists of key, values where the value for a given key is in the position
        just after the key e.g. [["foo", "bar", "head", 9]] => [{"foo": "bar", "head": 9}]

        Note: For backward compatibility, data can also be a dict.
        """
        if isinstance(data, dict):
            # for backward compatibility
            data = from_dict_to_key_value_list(data)

        parsed_dict = {}

        nested_model_list_fields = cls.get_nested_model_list_fields()
        nested_model_tuple_fields = cls.get_nested_model_tuple_fields()
        nested_model_fields = cls.get_nested_model_fields()

        for i in range(0, len(data), 2):
            key = from_bytes_to_str(data[i])
            field_type = cls._field_types.get(key)
            value = from_str_or_bytes_to_any(value=data[i + 1], field_type=field_type)

            if key in nested_model_list_fields and value is not None:
                value = deserialize_nested_model_list(
                    field_type=nested_model_list_fields[key], value=value
                )

            elif key in nested_model_tuple_fields and value is not None:
                value = deserialize_nested_model_tuple(
                    field_types=nested_model_tuple_fields[key], value=value
                )

            elif key in nested_model_fields and value is not None:
                value = deserialize_nested_model(
                    field_type=nested_model_fields[key], value=value
                )

            parsed_dict[key] = value

        return parsed_dict


def deserialize_nested_model_list(
    field_type: Type[AbstractModel], value: List[Any]
) -> List[AbstractModel]:
    """Deserializes a list of key values for the given field returning a list of nested models"""
    return [field_type(**field_type.deserialize_partially(item)) for item in value]


def deserialize_nested_model_tuple(
    field_types: Tuple[Any, ...], value: List[Any]
) -> Tuple[Any, ...]:
    """Deserializes a list of key values for the given field returning a tuple of nested models among others"""
    items = []
    for field_type, value in zip(field_types, value):
        if issubclass(field_type, AbstractModel) and value is not None:
            value = field_type(**field_type.deserialize_partially(value))
        items.append(value)

    return tuple(items)


def deserialize_nested_model(
    field_type: Type[AbstractModel], value: List[Any]
) -> AbstractModel:
    """Deserializes a list of key values for the given field returning the nested model"""
    return field_type(**field_type.deserialize_partially(value))



"""Module containing the mixin functionality for selecting"""
from typing import List, Any, Type, Union, Awaitable

from pydantic_redis.shared.model.prop_utils import (
    NESTED_MODEL_PREFIX,
    NESTED_MODEL_LIST_FIELD_PREFIX,
    NESTED_MODEL_TUPLE_FIELD_PREFIX,
    get_table_keys_regex,
    get_table_prefix,
)


from .base import AbstractModel


def get_select_fields(model: Type[AbstractModel], columns: List[str]) -> List[str]:
    """
    Gets the fields to be used for selecting HMAP fields in Redis
    It replaces any fields in `columns` that correspond to nested records with their
    `__` suffixed versions
    """
    fields = []
    nested_model_list_fields = model.get_nested_model_list_fields()
    nested_model_tuple_fields = model.get_nested_model_tuple_fields()
    nested_model_fields = model.get_nested_model_fields()

    for col in columns:

        if col in nested_model_fields:
            col = f"{NESTED_MODEL_PREFIX}{col}"
        elif col in nested_model_list_fields:
            col = f"{NESTED_MODEL_LIST_FIELD_PREFIX}{col}"
        elif col in nested_model_tuple_fields:
            col = f"{NESTED_MODEL_TUPLE_FIELD_PREFIX}{col}"

        fields.append(col)
    return fields


def select_all_fields_all_ids(
    model: Type[AbstractModel],
) -> Union[List[List[Any]], Awaitable[List[List[Any]]]]:
    """Selects all items in the database, returning all their fields"""
    table_keys_regex = get_table_keys_regex(model=model)
    args = [table_keys_regex]
    store = model.get_store()
    return store.select_all_fields_for_all_ids_script(args=args)


def select_all_fields_some_ids(
    model: Type[AbstractModel], ids: List[str]
) -> Union[List[List[Any]], Awaitable[List[List[Any]]]]:
    """Selects some items in the database, returning all their fields"""
    table_prefix = get_table_prefix(model=model)
    keys = [f"{table_prefix}{key}" for key in ids]
    store = model.get_store()
    return store.select_all_fields_for_some_ids_script(keys=keys)


def select_some_fields_all_ids(
    model: Type[AbstractModel], fields: List[str]
) -> Union[List[List[Any]], Awaitable[List[List[Any]]]]:
    """Selects all items in the database, returning only the specified fields"""
    table_keys_regex = get_table_keys_regex(model=model)
    columns = get_select_fields(model=model, columns=fields)
    args = [table_keys_regex, *columns]
    store = model.get_store()
    return store.select_some_fields_for_all_ids_script(args=args)


def select_some_fields_some_ids(
    model: Type[AbstractModel], fields: List[str], ids: List[str]
) -> Union[List[List[Any]], Awaitable[List[List[Any]]]]:
    """Selects some of items in the database, returning only the specified fields"""
    table_prefix = get_table_prefix(model=model)
    keys = [f"{table_prefix}{key}" for key in ids]
    columns = get_select_fields(model=model, columns=fields)
    store = model.get_store()
    return store.select_some_fields_for_all_ids_script(keys=keys, args=columns)


def parse_select_response(
    model: Type[AbstractModel], response: List[List], as_models: bool
):
    """
    Converts a list of lists of key-values into a list of models if `as_models` is true or leaves them as dicts
    with foreign keys replaced by model instances. The list is got from calling EVAL on Redis .

    EVAL returns a List of Lists of key, values where the value for a given key is in the position
    just after the key e.g. [["foo", "bar", "head", 9]] => [{"foo": "bar", "head": 9}]
    """
    if len(response) == 0:
        return None

    if as_models:
        return [
            model(**model.deserialize_partially(record))
            for record in response
            if record != []
        ]

    return [model.deserialize_partially(record) for record in response if record != []]



"""Module containing utils for getting properties of the Model"""

from typing import Type, Any

from .base import AbstractModel

NESTED_MODEL_PREFIX = "__"
NESTED_MODEL_LIST_FIELD_PREFIX = "___"
NESTED_MODEL_TUPLE_FIELD_PREFIX = "____"


def get_primary_key(model: Type[AbstractModel], primary_key_value: Any):
    """
    Returns the primary key value concatenated to the table name for uniqueness
    """
    return f"{get_table_prefix(model)}{primary_key_value}"


def get_table_prefix(model: Type[AbstractModel]):
    """
    Returns the prefix of the all the redis keys that are associated with this table
    """
    table_name = model.__name__.lower()
    return f"{table_name}_%&_"


def get_table_keys_regex(model: Type[AbstractModel]):
    """
    Returns the table name regex to get all keys that belong to this table
    """
    return f"{get_table_prefix(model)}*"


def get_table_index_key(model: Type[AbstractModel]):
    """Returns the key for the set in which the primary keys of the given table have been saved"""
    table_name = model.__name__.lower()
    return f"{table_name}__index"



"""Module containing the main config classes"""
from typing import Optional

from pydantic import BaseModel


class RedisConfig(BaseModel):
    host: str = "localhost"
    port: int = 6379
    db: int = 0
    password: Optional[str] = None
    ssl: bool = False
    encoding: Optional[str] = "utf-8"

    @property
    def redis_url(self) -> str:
        """Returns a redis url to connect to"""
        proto = "rediss" if self.ssl else "redis"
        if self.password is None:
            return f"{proto}://{self.host}:{self.port}/{self.db}"
        return f"{proto}://:{self.password}@{self.host}:{self.port}/{self.db}"

    class Config:
        orm_mode = True



"""Package containing the synchronous and thus default version of pydantic_redis"""
from .model import Model
from .store import Store
from ..shared.config import RedisConfig

__all__ = [Model, Store, RedisConfig]



"""Tests for benchmarks"""
import pytest

from test.conftest import (
    redis_store_fixture,
    Book,
    books,
    books_fixture,
    update_books_fixture,
    delete_books_fixture,
)


@pytest.mark.parametrize("store", redis_store_fixture)
def test_benchmark_bulk_insert(benchmark, store):
    """Benchmarks the bulk_insert operation"""
    benchmark(Book.insert, books)


@pytest.mark.parametrize("store, book", books_fixture)
def test_benchmark_single_insert(benchmark, store, book):
    """Benchmarks the single insert operation"""
    benchmark(Book.insert, book)


@pytest.mark.parametrize("store", redis_store_fixture)
def test_benchmark_select_default(benchmark, store):
    """Benchmarks the select default operation"""
    Book.insert(books)
    benchmark(Book.select)


@pytest.mark.parametrize("store", redis_store_fixture)
def test_benchmark_select_columns(benchmark, store):
    """Benchmarks the select columns operation"""
    Book.insert(books)
    benchmark(Book.select, columns=["title", "author", "in_stock"])


@pytest.mark.parametrize("store", redis_store_fixture)
def test_benchmark_select_some_items(benchmark, store):
    """Benchmarks the select some items operation"""
    Book.insert(books)
    ids = [book.title for book in books[:2]]
    benchmark(Book.select, ids=ids)


@pytest.mark.parametrize("store", redis_store_fixture)
def test_benchmark_select_columns_for_some_items(benchmark, store):
    """Benchmarks the select columns for some items only operation"""
    Book.insert(books)
    ids = [book.title for book in books[:2]]
    benchmark(Book.select, columns=["title", "author", "in_stock"], ids=ids)


@pytest.mark.parametrize("store, book", books_fixture)
def test_benchmark_select_columns_for_one_id(benchmark, store, book):
    """Benchmarks the select columns for one id operation"""
    Book.insert(books)
    benchmark(Book.select, columns=["title", "author", "in_stock"], ids=[book.title])


@pytest.mark.parametrize("store, book", books_fixture)
def test_benchmark_select_all_for_one_id(benchmark, store, book):
    """Benchmarks the select all columns for one id operation"""
    Book.insert(books)
    benchmark(Book.select, ids=[book.title])


@pytest.mark.parametrize("store, title, data", update_books_fixture)
def test_benchmark_update(benchmark, store, title, data):
    """Benchmarks the update operation"""
    Book.insert(books)
    benchmark(Book.update, title, data=data)


@pytest.mark.parametrize("store, title", delete_books_fixture)
def test_benchmark_delete(benchmark, store, title):
    """Benchmarks the delete operation"""
    Book.insert(books)
    benchmark(Book.delete, [title])


@pytest.mark.parametrize("store", redis_store_fixture)
def test_benchmark_bulk_delete(benchmark, store):
    """Benchmarks the bulk delete operation"""
    Book.insert(books)
    benchmark(Book.delete, [book.title for book in books])



"""Module containing the lua scripts used in select queries"""

SELECT_ALL_FIELDS_FOR_ALL_IDS_SCRIPT = """
local s_find = string.find
local s_gmatch = string.gmatch
local ipairs = ipairs
local table_insert = table.insert
local next = next
local redis_call = redis.call

local filtered = {}
local cursor = '0'

local function startswith(s, prefix)
    return s_find(s, prefix, 1, true) == 1
end

local function trim_dunder(s)
    return s:match '^_*(.-)$'
end

local function get_obj(id)
    local value = redis_call('HGETALL', id)

    for i, k in ipairs(value) do
        if not (i % 2 == 0) then
            if startswith(k, '___') or startswith(k, '____') then
                local nested = {}

                for v in s_gmatch(value[i + 1], '([^%[^,^%]^\"]+)') do
                    table_insert(nested, get_obj(v))
                end

                value[i + 1] = nested
                value[i] = trim_dunder(k)
            elseif startswith(k, '__') then
                value[i + 1] = get_obj(value[i + 1])
                value[i] = trim_dunder(k)
            end
        end
    end

    if next(value) == nil then
        return id
    end

    return value
end

repeat
    local result = redis_call('SCAN', cursor, 'MATCH', ARGV[1])
    for _, key in ipairs(result[2]) do
        if redis_call('TYPE', key).ok == 'hash' then
            table_insert(filtered, get_obj(key))
        end
    end
    cursor = result[1]
until (cursor == '0')
return filtered
"""

SELECT_ALL_FIELDS_FOR_SOME_IDS_SCRIPT = """
local s_find = string.find
local s_gmatch = string.gmatch
local ipairs = ipairs
local table_insert = table.insert
local next = next
local redis_call = redis.call

local result = {}

local function startswith(s, prefix)
    return s_find(s, prefix, 1, true) == 1
end

local function trim_dunder(s)
    return s:match '^_*(.-)$'
end

local function get_obj(id)
    local value = redis_call('HGETALL', id)

    for i, k in ipairs(value) do
        if not (i % 2 == 0) then
            if startswith(k, '___') or startswith(k, '____') then
                local nested = {}

                for v in s_gmatch(value[i + 1], '([^%[^,^%]^\"]+)') do
                    table_insert(nested, get_obj(v))
                end

                value[i + 1] = nested
                value[i] = trim_dunder(k)
            elseif startswith(k, '__') then
                value[i + 1] = get_obj(value[i + 1])
                value[i] = trim_dunder(k)
            end
        end
    end

    if next(value) == nil then
        return id
    end

    return value
end

for _, key in ipairs(KEYS) do
    local value = get_obj(key)
    if type(value) == 'table' then
        table_insert(result, value)
    end
end

return result
"""

SELECT_SOME_FIELDS_FOR_ALL_IDS_SCRIPT = """
local s_find = string.find
local s_gmatch = string.gmatch
local ipairs = ipairs
local table_insert = table.insert
local next = next
local redis_call = redis.call
local table_unpack = table.unpack or unpack

local filtered = {}
local cursor = '0'
local columns = {}

local function startswith(s, prefix)
    return s_find(s, prefix, 1, true) == 1
end

local function trim_dunder(s)
    return s:match '^_*(.-)$'
end

local function get_obj(id)
    local value = redis_call('HGETALL', id)

    for i, k in ipairs(value) do
        if not (i % 2 == 0) then
            if startswith(k, '___') or startswith(k, '____') then
                local nested = {}

                for v in s_gmatch(value[i + 1], '([^%[^,^%]^\"]+)') do
                    table_insert(nested, get_obj(v))
                end

                value[i + 1] = nested
                value[i] = trim_dunder(k)
            elseif startswith(k, '__') then
                value[i + 1] = get_obj(value[i + 1])
                value[i] = trim_dunder(k)
            end
        end
    end

    if next(value) == nil then
        return id
    end

    return value
end

for i, k in ipairs(ARGV) do
    if i > 1 then
        table_insert(columns, k)
    end
end

repeat
    local result = redis_call('SCAN', cursor, 'MATCH', ARGV[1])
    for _, key in ipairs(result[2]) do
        if redis_call('TYPE', key).ok == 'hash' then
            local data = redis_call('HMGET', key, table_unpack(columns))
            local parsed_data = {}

            for i, v in ipairs(data) do
                table_insert(parsed_data, trim_dunder(columns[i]))
                table_insert(parsed_data, get_obj(v))
            end

            table_insert(filtered, parsed_data)
        end
    end
    cursor = result[1]
until (cursor == '0')
return filtered
"""

SELECT_SOME_FIELDS_FOR_SOME_IDS_SCRIPT = """
local s_find = string.find
local s_gmatch = string.gmatch
local ipairs = ipairs
local table_insert = table.insert
local next = next
local redis_call = redis.call
local table_unpack = table.unpack or unpack

local result = {}
local columns = {}

local function startswith(s, prefix)
    return s_find(s, prefix, 1, true) == 1
end

local function trim_dunder(s)
    return s:match '^_*(.-)$'
end

local function get_obj(id)
    local value = redis_call('HGETALL', id)

    for i, k in ipairs(value) do
        if not (i % 2 == 0) then
            if startswith(k, '___') or startswith(k, '____') then
                local nested = {}

                for v in s_gmatch(value[i + 1], '([^%[^,^%]^\"]+)') do
                    table_insert(nested, get_obj(v))
                end

                value[i + 1] = nested
                value[i] = trim_dunder(k)
            elseif startswith(k, '__') then
                value[i + 1] = get_obj(value[i + 1])
                value[i] = trim_dunder(k)
            end
        end
    end

    if next(value) == nil then
        return id
    end

    return value
end

for _, k in ipairs(ARGV) do
    table_insert(columns, k)
end

for _, key in ipairs(KEYS) do
    local data = redis_call('HMGET', key, table_unpack(columns))
    local parsed_data = {}

    for i, v in ipairs(data) do
        if v then
            table_insert(parsed_data, trim_dunder(columns[i]))
            table_insert(parsed_data, get_obj(v))
        end
    end

    table_insert(result, parsed_data)
end
return result
"""



"""Module containing the mixin for insert functionality in model"""
from typing import Union, Optional, Any, Dict, Tuple, List, Type

from redis.asyncio.client import Pipeline as AioPipeline
from redis.client import Pipeline

from .prop_utils import (
    get_primary_key,
    get_table_index_key,
    NESTED_MODEL_PREFIX,
    NESTED_MODEL_LIST_FIELD_PREFIX,
    NESTED_MODEL_TUPLE_FIELD_PREFIX,
)

from .base import AbstractModel


def insert_on_pipeline(
    model: Type[AbstractModel],
    pipeline: Union[Pipeline, AioPipeline],
    _id: Optional[Any],
    record: Union[AbstractModel, Dict[str, Any]],
    life_span: Optional[Union[float, int]] = None,
) -> Any:
    """
    Creates insert commands for the given record on the given pipeline but does not execute
    thus the data is not yet persisted in redis
    Returns the key of the created item
    """
    key = _id if _id is not None else getattr(record, model.get_primary_key_field())
    data = _get_serializable_dict(
        model=model, pipeline=pipeline, record=record, life_span=life_span
    )
    name = get_primary_key(model=model, primary_key_value=key)
    mapping = model.serialize_partially(data)
    pipeline.hset(name=name, mapping=mapping)

    if life_span is not None:
        pipeline.expire(name=name, time=life_span)
    # save the primary key in an index
    table_index_key = get_table_index_key(model)
    pipeline.sadd(table_index_key, name)
    if life_span is not None:
        pipeline.expire(table_index_key, time=life_span)

    return name


def _get_serializable_dict(
    model: Type[AbstractModel],
    pipeline: Union[Pipeline, AioPipeline],
    record: Union[AbstractModel, Dict[str, Any]],
    life_span: Optional[Union[float, int]] = None,
) -> Dict[str, Any]:
    """
    Returns a dictionary that can be serialized.
    A few cleanups it does include:
      - Upserting any nested records in `record`
      - Replacing the keys of nested records with their NESTED_MODEL_PREFIX suffixed versions
        e.g. `__author` instead of author
      - Replacing the keys of lists of nested records with their NESTED_MODEL_LIST_FIELD_PREFIX suffixed versions
        e.g. `__%&l_author` instead of author
      - Replacing the keys of tuples of nested records with their NESTED_MODEL_TUPLE_FIELD_PREFIX suffixed versions
        e.g. `__%&l_author` instead of author
      - Replacing the values of nested records with their foreign keys
    """
    data = record.items() if isinstance(record, dict) else record
    new_data = {}

    nested_model_list_fields = model.get_nested_model_list_fields()
    nested_model_tuple_fields = model.get_nested_model_tuple_fields()
    nested_model_fields = model.get_nested_model_fields()

    for k, v in data:
        key, value = k, v

        if key in nested_model_list_fields:
            key, value = _serialize_nested_model_list_field(
                key=key, value=value, pipeline=pipeline, life_span=life_span
            )
        elif key in nested_model_tuple_fields:
            key, value = _serialize_nested_model_tuple_field(
                key=key,
                value=value,
                pipeline=pipeline,
                life_span=life_span,
                tuple_fields=nested_model_tuple_fields,
            )
        elif key in nested_model_fields:
            key, value = _serialize_nested_model_field(
                key=key, value=value, pipeline=pipeline, life_span=life_span
            )

        new_data[key] = value
    return new_data


def _serialize_nested_model_tuple_field(
    key: str,
    value: Tuple[AbstractModel],
    pipeline: Union[Pipeline, AioPipeline],
    life_span: Optional[Union[float, int]],
    tuple_fields: Dict[str, Tuple[Any, ...]],
) -> Tuple[str, List[Any]]:
    """Serializes a key-value pair for a field that has a tuple of nested models"""
    try:
        field_types = tuple_fields.get(key, ())
        value = [
            insert_on_pipeline(
                model=field_type,
                _id=None,
                pipeline=pipeline,
                record=item,
                life_span=life_span,
            )
            # FIXME: The reference to AbstractModel here. Is it right?
            if issubclass(field_type, AbstractModel) else item
            for field_type, item in zip(field_types, value)
        ]
        key = f"{NESTED_MODEL_TUPLE_FIELD_PREFIX}{key}"
    except TypeError:
        # In case the value is None, just ignore
        pass

    return key, value


def _serialize_nested_model_list_field(
    key: str,
    value: List[AbstractModel],
    pipeline: Union[Pipeline, AioPipeline],
    life_span: Optional[Union[float, int]],
) -> Tuple[str, List[Any]]:
    """Serializes a key-value pair for a field that has a list of nested models"""
    try:
        value = [
            insert_on_pipeline(
                model=item.__class__,
                _id=None,
                pipeline=pipeline,
                record=item,
                life_span=life_span,
            )
            for item in value
        ]
        key = f"{NESTED_MODEL_LIST_FIELD_PREFIX}{key}"
    except TypeError:
        # In case the value is None, just ignore
        pass

    return key, value


def _serialize_nested_model_field(
    key: str,
    value: AbstractModel,
    pipeline: Union[Pipeline, AioPipeline],
    life_span: Optional[Union[float, int]],
) -> Tuple[str, str]:
    """Serializes a key-value pair for a field that has a nested model"""
    try:
        value = insert_on_pipeline(
            model=value.__class__,
            _id=None,
            pipeline=pipeline,
            record=value,
            life_span=life_span,
        )
        key = f"{NESTED_MODEL_PREFIX}{key}"
    except TypeError:
        # In case the value is None, just ignore
        pass

    return key, value



"""Module containing the model classes"""
from typing import Optional, List, Any, Union, Dict, Tuple, Type

import redis.asyncio

from pydantic_redis.shared.model import AbstractModel
from pydantic_redis.shared.model.insert_utils import insert_on_pipeline
from pydantic_redis.shared.model.prop_utils import get_primary_key, get_table_index_key
from pydantic_redis.shared.model.select_utils import (
    select_all_fields_all_ids,
    select_all_fields_some_ids,
    select_some_fields_all_ids,
    select_some_fields_some_ids,
    parse_select_response,
)

from .store import Store


class Model(AbstractModel):
    """
    The section in the store that saves rows of the same kind
    """

    _store: Store

    @classmethod
    async def insert(
        cls,
        data: Union[List[AbstractModel], AbstractModel],
        life_span_seconds: Optional[float] = None,
    ):
        """
        Inserts a given row or sets of rows into the table
        """
        store = cls.get_store()
        life_span = (
            life_span_seconds
            if life_span_seconds is not None
            else store.life_span_in_seconds
        )

        async with store.redis_store.pipeline(transaction=True) as pipeline:
            data_list = []

            if isinstance(data, list):
                data_list = data
            elif isinstance(data, AbstractModel):
                data_list = [data]

            for record in data_list:
                insert_on_pipeline(
                    model=cls,
                    _id=None,
                    pipeline=pipeline,
                    record=record,
                    life_span=life_span,
                )

            return await pipeline.execute()

    @classmethod
    async def update(
        cls, _id: Any, data: Dict[str, Any], life_span_seconds: Optional[float] = None
    ):
        """
        Updates a given row or sets of rows in the table
        """
        store = cls.get_store()
        life_span = (
            life_span_seconds
            if life_span_seconds is not None
            else store.life_span_in_seconds
        )
        async with store.redis_store.pipeline(transaction=True) as pipeline:
            if isinstance(data, dict):
                insert_on_pipeline(
                    model=cls,
                    _id=_id,
                    pipeline=pipeline,
                    record=data,
                    life_span=life_span,
                )

            return await pipeline.execute()

    @classmethod
    async def delete(cls, ids: Union[Any, List[Any]]):
        """
        deletes a given row or sets of rows in the table
        """
        store = cls.get_store()

        async with store.redis_store.pipeline() as pipeline:
            primary_keys = []

            if isinstance(ids, list):
                primary_keys = ids
            elif ids is not None:
                primary_keys = [ids]

            names = [
                get_primary_key(model=cls, primary_key_value=primary_key_value)
                for primary_key_value in primary_keys
            ]
            pipeline.delete(*names)
            # remove the primary keys from the index
            table_index_key = get_table_index_key(model=cls)
            pipeline.srem(table_index_key, *names)
            return await pipeline.execute()

    @classmethod
    async def select(
        cls,
        columns: Optional[List[str]] = None,
        ids: Optional[List[Any]] = None,
        **kwargs,
    ):
        """
        Selects given rows or sets of rows in the table
        """
        if columns is None and ids is None:
            response = await select_all_fields_all_ids(model=cls)

        elif columns is None and isinstance(ids, list):
            response = await select_all_fields_some_ids(model=cls, ids=ids)

        elif isinstance(columns, list) and ids is None:
            response = await select_some_fields_all_ids(model=cls, fields=columns)

        elif isinstance(columns, list) and isinstance(ids, list):
            response = await select_some_fields_some_ids(
                model=cls, fields=columns, ids=ids
            )

        else:
            raise ValueError(
                f"columns {columns}, ids: {ids} should be either None or lists"
            )

        return parse_select_response(
            model=cls, response=response, as_models=(columns is None)
        )



"""Module containing common base store functionality"""
from typing import Optional, Union, Type, Dict, Any

from redis import Redis
from redis.asyncio import Redis as AioRedis
from pydantic import BaseModel
from redis.commands.core import Script, AsyncScript

from .config import RedisConfig
from .lua_scripts import (
    SELECT_ALL_FIELDS_FOR_ALL_IDS_SCRIPT,
    SELECT_ALL_FIELDS_FOR_SOME_IDS_SCRIPT,
    SELECT_SOME_FIELDS_FOR_ALL_IDS_SCRIPT,
    SELECT_SOME_FIELDS_FOR_SOME_IDS_SCRIPT,
)

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .model import AbstractModel


class AbstractStore(BaseModel):
    """
    An abstract class of a store
    """

    name: str
    redis_config: RedisConfig
    redis_store: Optional[Union[Redis, AioRedis]] = None
    life_span_in_seconds: Optional[int] = None
    select_all_fields_for_all_ids_script: Optional[Union[AsyncScript, Script]] = None
    select_all_fields_for_some_ids_script: Optional[Union[AsyncScript, Script]] = None
    select_some_fields_for_all_ids_script: Optional[Union[AsyncScript, Script]] = None
    select_some_fields_for_some_ids_script: Optional[Union[AsyncScript, Script]] = None
    models: Dict[str, Type["AbstractModel"]] = {}

    class Config:
        arbitrary_types_allowed = True
        orm_mode = True

    def __init__(
        self,
        name: str,
        redis_config: RedisConfig,
        redis_store: Optional[Union[Redis, AioRedis]] = None,
        life_span_in_seconds: Optional[int] = None,
        **data: Any,
    ):
        super().__init__(
            name=name,
            redis_config=redis_config,
            redis_store=redis_store,
            life_span_in_seconds=life_span_in_seconds,
            **data,
        )

        self.redis_store = self._connect_to_redis()
        self._register_lua_scripts()

    def _connect_to_redis(self) -> Union[Redis, AioRedis]:
        """Connects the store to redis, returning a proper connection"""
        raise NotImplementedError("implement _connect_to_redis first")

    def _register_lua_scripts(self):
        """Registers the lua scripts for this redis instance"""
        self.select_all_fields_for_all_ids_script = self.redis_store.register_script(
            SELECT_ALL_FIELDS_FOR_ALL_IDS_SCRIPT
        )
        self.select_all_fields_for_some_ids_script = self.redis_store.register_script(
            SELECT_ALL_FIELDS_FOR_SOME_IDS_SCRIPT
        )
        self.select_some_fields_for_all_ids_script = self.redis_store.register_script(
            SELECT_SOME_FIELDS_FOR_ALL_IDS_SCRIPT
        )
        self.select_some_fields_for_some_ids_script = self.redis_store.register_script(
            SELECT_SOME_FIELDS_FOR_SOME_IDS_SCRIPT
        )

    def register_model(self, model_class: Type["AbstractModel"]):
        """Registers the model to this store"""
        if not isinstance(model_class.get_primary_key_field(), str):
            raise NotImplementedError(
                f"{model_class.__name__} should have a _primary_key_field"
            )

        model_class._store = self
        model_class.initialize()
        self.models[model_class.__name__.lower()] = model_class

    def model(self, name: str) -> Type["AbstractModel"]:
        """Gets a model by name: case insensitive"""
        return self.models[name.lower()]



"""Module containing the store classes for sync io"""
from typing import Dict, Type, TYPE_CHECKING

import redis

from ..shared.store import AbstractStore

if TYPE_CHECKING:
    from .model import Model


class Store(AbstractStore):
    """
    A store that allows a declarative way of querying for data in redis
    """

    models: Dict[str, Type["Model"]] = {}

    def _connect_to_redis(self) -> redis.Redis:
        """Connects the store to redis, returning a proper connection"""
        return redis.from_url(
            self.redis_config.redis_url,
            encoding=self.redis_config.encoding,
            decode_responses=True,
        )



"""Module containing common utilities"""
import typing
from typing import Any, Tuple, Optional, Union, Dict, Callable, Type, List

import orjson


def strip_leading(word: str, substring: str) -> str:
    """
    Strips the leading substring if it exists.
    This is contrary to rstrip which can looks at removes each character in the substring
    """
    if word.startswith(substring):
        return word[len(substring) :]
    return word


def typing_get_args(v: Any) -> Tuple[Any, ...]:
    """Gets the __args__ of the annotations of a given typing"""
    try:
        return typing.get_args(v)
    except AttributeError:
        return getattr(v, "__args__", ()) if v is not typing.Generic else typing.Generic


def typing_get_origin(v: Any) -> Optional[Any]:
    """Gets the __origin__ of the annotations of a given typing"""
    try:
        return typing.get_origin(v)
    except AttributeError:
        return getattr(v, "__origin__", None)


def from_bytes_to_str(value: Union[str, bytes]) -> str:
    """Converts bytes to str"""
    if isinstance(value, bytes):
        return str(value, "utf-8")
    return value


def from_str_or_bytes_to_any(value: Any, field_type: Type) -> Any:
    """Converts str or bytes to arbitrary data"""
    if isinstance(value, (bytes, bytearray, memoryview)):
        return orjson.loads(value)
    elif isinstance(value, str) and field_type != str:
        return orjson.loads(value)
    return value


def from_any_to_str_or_bytes(value: Any) -> Union[str, bytes]:
    """Converts arbitrary data into str or bytes"""
    if isinstance(value, str):
        return value
    return orjson.dumps(value, default=default_json_dump)


def default_json_dump(obj):
    """Default JSON dump for orjson"""
    if hasattr(obj, "json") and isinstance(obj.json, Callable):
        return obj.json()
    elif isinstance(obj, set):
        # Set does not exist in JSON.
        # It's fine to use list instead, it becomes a Set when deserializing.
        return list(obj)


def from_dict_to_key_value_list(data: Dict[str, Any]) -> List[Any]:
    """Converts dict to flattened list of key, values where the value after the key"""
    parsed_list = []

    for k, v in data.items():
        parsed_list.append(k)
        parsed_list.append(v)

    return parsed_list



"""Tests for the redis orm"""

from typing import Dict, Any

import pytest

from pydantic_redis.shared.model.prop_utils import NESTED_MODEL_PREFIX
from pydantic_redis.shared.utils import strip_leading
from pydantic_redis.asyncio import Model, RedisConfig
from test.conftest import (
    async_redis_store_fixture,
    AsyncBook,
    async_books,
    AsyncAuthor,
    async_authors,
    AsyncLibrary,
)


def test_redis_config_redis_url():
    password = "password"
    config_with_no_pass = RedisConfig()
    config_with_ssl = RedisConfig(ssl=True)
    config_with_pass = RedisConfig(password=password)
    config_with_pass_ssl = RedisConfig(ssl=True, password=password)

    assert config_with_no_pass.redis_url == "redis://localhost:6379/0"
    assert config_with_ssl.redis_url == "rediss://localhost:6379/0"
    assert config_with_pass.redis_url == f"redis://:{password}@localhost:6379/0"
    assert config_with_pass_ssl.redis_url == f"rediss://:{password}@localhost:6379/0"


def test_register_model_without_primary_key(async_redis_store):
    """Throws error when a model without the _primary_key_field class variable set is registered"""

    class ModelWithoutPrimaryKey(Model):
        title: str

    with pytest.raises(AttributeError, match=r"_primary_key_field"):
        async_redis_store.register_model(ModelWithoutPrimaryKey)

    ModelWithoutPrimaryKey._primary_key_field = None

    with pytest.raises(Exception, match=r"should have a _primary_key_field"):
        async_redis_store.register_model(ModelWithoutPrimaryKey)


def test_store_model(async_redis_store):
    """Tests the model method in store"""
    assert async_redis_store.model("AsyncBook") == AsyncBook

    with pytest.raises(KeyError):
        async_redis_store.model("Notabook")


@pytest.mark.asyncio
@pytest.mark.parametrize("store", async_redis_store_fixture)
async def test_bulk_insert(store):
    """Providing a list of Model instances to the insert method inserts the records in redis"""
    book_keys = [f"asyncbook_%&_{book.title}" for book in async_books]
    keys = book_keys + [
        f"asyncauthor_%&_{author.name}" for author in async_authors.values()
    ]
    await store.redis_store.delete(*keys)

    for key in keys:
        item_in_redis = await store.redis_store.hgetall(name=key)
        assert item_in_redis == {}

    await AsyncBook.insert(async_books)

    async with store.redis_store.pipeline() as pipeline:
        for book_key in book_keys:
            pipeline.hgetall(name=book_key)
        async_books_in_redis = await pipeline.execute()

    async_books_in_redis_as_models = [
        await __deserialize_book_data(book) for book in async_books_in_redis
    ]
    assert async_books == async_books_in_redis_as_models


@pytest.mark.asyncio
@pytest.mark.parametrize("store", async_redis_store_fixture)
async def test_bulk_nested_insert(store):
    """Providing a list of Model instances to the insert method also upserts their nested records in redis"""
    book_keys = [f"asyncbook_%&_{book.title}" for book in async_books]
    author_keys = [f"asyncauthor_%&_{author.name}" for author in async_authors.values()]
    keys = book_keys + author_keys

    for key in keys:
        item_in_redis = await store.redis_store.hgetall(name=key)
        assert item_in_redis == {}

    await AsyncBook.insert(async_books)

    async with store.redis_store.pipeline() as pipeline:
        for key in author_keys:
            pipeline.hgetall(name=key)
        async_authors_in_redis = await pipeline.execute()
    async_authors_in_redis_as_models = sorted(
        [
            AsyncAuthor(**AsyncAuthor.deserialize_partially(author))
            for author in async_authors_in_redis
        ],
        key=lambda x: x.name,
    )
    expected = sorted(async_authors.values(), key=lambda x: x.name)
    assert expected == async_authors_in_redis_as_models


@pytest.mark.asyncio
@pytest.mark.parametrize("store", async_redis_store_fixture)
async def test_insert_single(store):
    """
    Providing a single Model instance inserts that record in redis
    """
    key = f"asyncbook_%&_{async_books[0].title}"
    book = await store.redis_store.hgetall(name=key)
    assert book == {}

    await AsyncBook.insert(async_books[0])

    book = await store.redis_store.hgetall(name=key)
    book_as_model = await __deserialize_book_data(book)
    assert async_books[0] == book_as_model


@pytest.mark.asyncio
@pytest.mark.parametrize("store", async_redis_store_fixture)
async def test_insert_single_nested(store):
    """
    Providing a single Model instance upserts also any nested model into redis
    """
    key = f"asyncauthor_%&_{async_books[0].author.name}"
    author = await store.redis_store.hgetall(name=key)
    assert author == {}

    await AsyncBook.insert(async_books[0])

    author = await store.redis_store.hgetall(name=key)
    author_as_model = AsyncAuthor(**AsyncAuthor.deserialize_partially(author))
    assert async_books[0].author == author_as_model


@pytest.mark.asyncio
@pytest.mark.parametrize("store", async_redis_store_fixture)
async def test_update_nested_list_of_models(store):
    data = [
        AsyncLibrary(name="Babel AsyncLibrary", address="In a book", books=async_books)
    ]
    await AsyncLibrary.insert(data)
    # the list of nested models is automatically inserted
    got = sorted(await AsyncBook.select(), key=lambda x: x.title)
    expected = sorted(async_books, key=lambda x: x.title)
    assert expected == got

    got = sorted(await AsyncLibrary.select(), key=lambda x: x.name)
    expected = sorted(data, key=lambda x: x.name)
    assert got == expected


@pytest.mark.asyncio
@pytest.mark.parametrize("store", async_redis_store_fixture)
async def test_update_optional_nested_list_of_models(store):
    data = [
        AsyncLibrary(name="Babel AsyncLibrary", address="In a book", lost=async_books)
    ]
    await AsyncLibrary.insert(data)
    # the list of nested models is automatically inserted
    got = sorted(await AsyncBook.select(), key=lambda x: x.title)
    expected = sorted(async_books, key=lambda x: x.title)
    assert expected == got

    got = sorted(await AsyncLibrary.select(), key=lambda x: x.name)
    expected = sorted(data, key=lambda x: x.name)
    assert got == expected


@pytest.mark.asyncio
@pytest.mark.parametrize("store", async_redis_store_fixture)
async def test_update_nested_tuple_of_models(store):
    jane = async_authors["jane"]
    new_stuff = (async_books[0], jane, async_books[1], 8)
    data = [AsyncLibrary(name="Babel AsyncLibrary", address="In a book", new=new_stuff)]
    await AsyncLibrary.insert(data)
    # the tuple of nested models is automatically inserted
    got = sorted(await AsyncBook.select(), key=lambda x: x.title)
    expected = sorted([async_books[0], async_books[1]], key=lambda x: x.title)
    assert expected == got

    got = sorted(await AsyncAuthor.select(), key=lambda x: x.name)
    expected = sorted([async_books[0].author, jane], key=lambda x: x.name)
    assert expected == got

    got = sorted(await AsyncLibrary.select(), key=lambda x: x.name)
    expected = sorted(data, key=lambda x: x.name)
    assert got == expected


@pytest.mark.asyncio
@pytest.mark.parametrize("store", async_redis_store_fixture)
async def test_update_optional_nested_tuple_of_models(store):
    popular_async_books = (async_books[0], async_books[2])
    data = [
        AsyncLibrary(
            name="Babel AsyncLibrary", address="In a book", popular=popular_async_books
        )
    ]
    await AsyncLibrary.insert(data)
    # the tuple of nested models is automatically inserted
    got = sorted(await AsyncBook.select(), key=lambda x: x.title)
    expected = sorted(popular_async_books, key=lambda x: x.title)
    assert expected == got

    got = sorted(await AsyncLibrary.select(), key=lambda x: x.name)
    expected = sorted(data, key=lambda x: x.name)
    assert got == expected


@pytest.mark.asyncio
@pytest.mark.parametrize("store", async_redis_store_fixture)
async def test_select_default(store):
    """Selecting without arguments returns all the book models"""
    await AsyncBook.insert(async_books)
    response = await AsyncBook.select()
    sorted_async_books = sorted(async_books, key=lambda x: x.title)
    sorted_response = sorted(response, key=lambda x: x.title)
    assert sorted_async_books == sorted_response


@pytest.mark.asyncio
@pytest.mark.parametrize("store", async_redis_store_fixture)
async def test_select_some_columns(store):
    """
    Selecting some columns returns a list of dictionaries of all async_books models with only those columns
    """
    await AsyncBook.insert(async_books)
    async_books_dict = {book.title: book for book in async_books}
    columns = ["title", "author", "in_stock"]
    response = await AsyncBook.select(columns=["title", "author", "in_stock"])
    response_dict = {book["title"]: book for book in response}

    for title, book in async_books_dict.items():
        book_in_response = response_dict[title]
        assert isinstance(book_in_response, dict)
        assert sorted(book_in_response.keys()) == sorted(columns)

        for column in columns:
            if column == "author":
                assert book_in_response[column] == getattr(book, column)
            else:
                assert f"{book_in_response[column]}" == f"{getattr(book, column)}"


@pytest.mark.asyncio
@pytest.mark.parametrize("store", async_redis_store_fixture)
async def test_select_some_ids(store):
    """
    Selecting some ids returns only those elements with the given ids
    """
    await AsyncBook.insert(async_books)
    ids = [book.title for book in async_books[:2]]
    response = await AsyncBook.select(ids=ids)
    assert response == async_books[:2]


@pytest.mark.asyncio
@pytest.mark.parametrize("store", async_redis_store_fixture)
async def test_update(store):
    """
    Updating an item of a given primary key updates it in redis
    """
    await AsyncBook.insert(async_books)
    title = async_books[0].title
    new_in_stock = not async_books[0].in_stock
    new_author = AsyncAuthor(name="John Doe", active_years=(2000, 2009))
    book_key = f"asyncbook_%&_{title}"
    new_author_key = f"asyncauthor_%&_{new_author.name}"
    old_book_data = await store.redis_store.hgetall(name=book_key)
    old_book = await __deserialize_book_data(old_book_data)
    assert old_book == async_books[0]
    assert old_book.author != new_author

    await AsyncBook.update(
        _id=title, data={"author": new_author, "in_stock": new_in_stock}
    )

    book_data = await store.redis_store.hgetall(name=book_key)
    book = await __deserialize_book_data(book_data)
    author_data = await store.redis_store.hgetall(name=new_author_key)
    author = AsyncAuthor(**AsyncAuthor.deserialize_partially(author_data))
    assert book.author == new_author
    assert author == new_author
    assert book.title == old_book.title
    assert book.in_stock == new_in_stock
    assert book.published_on == old_book.published_on


@pytest.mark.asyncio
@pytest.mark.parametrize("store", async_redis_store_fixture)
async def test_update_nested_model(store):
    """
    Updating a nested model, without changing its primary key, also updates it its collection in redis
    """
    await AsyncBook.insert(async_books)

    new_in_stock = not async_books[0].in_stock
    updated_author = AsyncAuthor(**async_books[0].author.dict())
    updated_author.active_years = (2020, 2045)
    book_key = f"asyncbook_%&_{async_books[0].title}"
    author_key = f"asyncauthor_%&_{updated_author.name}"

    old_author_data = await store.redis_store.hgetall(name=author_key)
    old_author = AsyncAuthor(**AsyncAuthor.deserialize_partially(old_author_data))
    old_book_data = await store.redis_store.hgetall(name=book_key)
    old_book = await __deserialize_book_data(old_book_data)
    assert old_book == async_books[0]
    assert old_author == async_books[0].author
    assert old_author != updated_author

    await AsyncBook.update(
        _id=async_books[0].title,
        data={"author": updated_author, "in_stock": new_in_stock},
    )

    book_data = await store.redis_store.hgetall(name=book_key)
    book = await __deserialize_book_data(book_data)
    author_data = await store.redis_store.hgetall(name=author_key)
    author = AsyncAuthor(**AsyncAuthor.deserialize_partially(author_data))
    assert book.author == updated_author
    assert author == updated_author
    assert book.title == old_book.title
    assert book.in_stock == new_in_stock
    assert book.published_on == old_book.published_on


@pytest.mark.asyncio
@pytest.mark.parametrize("store", async_redis_store_fixture)
async def test_delete_multiple(store):
    """
    Providing a list of ids to the delete function will remove the items from redis,
    but leave the nested models intact
    """
    await AsyncBook.insert(async_books)
    async_books_to_delete = async_books[:2]
    async_books_left_in_db = async_books[2:]

    ids_to_delete = [book.title for book in async_books_to_delete]
    ids_to_leave_intact = [book.title for book in async_books_left_in_db]

    keys_to_delete = [f"asyncbook_%&_{_id}" for _id in ids_to_delete]
    book_keys_to_leave_intact = [f"asyncbook_%&_{_id}" for _id in ids_to_leave_intact]
    author_keys_to_leave_intact = [
        f"asyncauthor_%&_{author.name}" for author in async_authors.values()
    ]

    await AsyncBook.delete(ids=ids_to_delete)

    for key in keys_to_delete:
        deleted_book_in_redis = await store.redis_store.hgetall(name=key)
        assert deleted_book_in_redis == {}

    async with store.redis_store.pipeline() as pipeline:
        for key in book_keys_to_leave_intact:
            pipeline.hgetall(name=key)
        async_books_in_redis = await pipeline.execute()
        async_books_in_redis_as_models = [
            await __deserialize_book_data(book) for book in async_books_in_redis
        ]
        assert async_books_left_in_db == async_books_in_redis_as_models

        for key in author_keys_to_leave_intact:
            pipeline.hgetall(name=key)
        async_authors_in_redis = await pipeline.execute()
        async_authors_in_redis_as_models = sorted(
            [
                AsyncAuthor(**AsyncAuthor.deserialize_partially(author))
                for author in async_authors_in_redis
            ],
            key=lambda x: x.name,
        )
        expected = sorted(async_authors.values(), key=lambda x: x.name)
        assert expected == async_authors_in_redis_as_models


async def __deserialize_book_data(raw_book_data: Dict[str, Any]) -> AsyncBook:
    """Deserializes the raw book data returning a book instance"""
    author_id = raw_book_data.pop(f"{NESTED_MODEL_PREFIX}author")
    author_id = strip_leading(author_id, "asyncauthor_%&_")

    data = AsyncBook.deserialize_partially(raw_book_data)

    data["author"] = (await AsyncAuthor.select(ids=[author_id]))[0]
    return AsyncBook(**data)



"""Tests for the redis orm"""

from typing import Dict, Any

import pytest

from pydantic_redis.shared.config import RedisConfig
from pydantic_redis.shared.model.prop_utils import NESTED_MODEL_PREFIX
from pydantic_redis.shared.utils import strip_leading
from pydantic_redis.syncio.model import Model
from test.conftest import (
    redis_store_fixture,
    Book,
    books,
    Author,
    authors,
    Library,
)


def test_redis_config_redis_url():
    password = "password"
    config_with_no_pass = RedisConfig()
    config_with_ssl = RedisConfig(ssl=True)
    config_with_pass = RedisConfig(password=password)
    config_with_pass_ssl = RedisConfig(ssl=True, password=password)

    assert config_with_no_pass.redis_url == "redis://localhost:6379/0"
    assert config_with_ssl.redis_url == "rediss://localhost:6379/0"
    assert config_with_pass.redis_url == f"redis://:{password}@localhost:6379/0"
    assert config_with_pass_ssl.redis_url == f"rediss://:{password}@localhost:6379/0"


def test_register_model_without_primary_key(redis_store):
    """Throws error when a model without the _primary_key_field class variable set is registered"""

    class ModelWithoutPrimaryKey(Model):
        title: str

    with pytest.raises(AttributeError, match=r"_primary_key_field"):
        redis_store.register_model(ModelWithoutPrimaryKey)

    ModelWithoutPrimaryKey._primary_key_field = None

    with pytest.raises(Exception, match=r"should have a _primary_key_field"):
        redis_store.register_model(ModelWithoutPrimaryKey)


def test_store_model(redis_store):
    """Tests the model method in store"""
    assert redis_store.model("Book") == Book

    with pytest.raises(KeyError):
        redis_store.model("Notabook")


@pytest.mark.parametrize("store", redis_store_fixture)
def test_bulk_insert(store):
    """Providing a list of Model instances to the insert method inserts the records in redis"""
    book_keys = [f"book_%&_{book.title}" for book in books]
    keys = book_keys + [f"author_%&_{author.name}" for author in authors.values()]
    store.redis_store.delete(*keys)

    for key in keys:
        item_in_redis = store.redis_store.hgetall(name=key)
        assert item_in_redis == {}

    Book.insert(books)

    with store.redis_store.pipeline() as pipeline:
        for book_key in book_keys:
            pipeline.hgetall(name=book_key)
        books_in_redis = pipeline.execute()
    books_in_redis_as_models = [
        __deserialize_book_data(book) for book in books_in_redis
    ]
    assert books == books_in_redis_as_models


@pytest.mark.parametrize("store", redis_store_fixture)
def test_bulk_nested_insert(store):
    """Providing a list of Model instances to the insert method also upserts their nested records in redis"""
    book_keys = [f"book_%&_{book.title}" for book in books]
    author_keys = [f"author_%&_{author.name}" for author in authors.values()]
    keys = book_keys + author_keys

    for key in keys:
        item_in_redis = store.redis_store.hgetall(name=key)
        assert item_in_redis == {}

    Book.insert(books)

    with store.redis_store.pipeline() as pipeline:
        for key in author_keys:
            pipeline.hgetall(name=key)
        authors_in_redis = pipeline.execute()
    authors_in_redis_as_models = sorted(
        [Author(**Author.deserialize_partially(author)) for author in authors_in_redis],
        key=lambda x: x.name,
    )
    expected = sorted(authors.values(), key=lambda x: x.name)
    assert expected == authors_in_redis_as_models


@pytest.mark.parametrize("store", redis_store_fixture)
def test_insert_single(store):
    """
    Providing a single Model instance inserts that record in redis
    """
    key = f"book_%&_{books[0].title}"
    book = store.redis_store.hgetall(name=key)
    assert book == {}

    Book.insert(books[0])

    book = store.redis_store.hgetall(name=key)
    book_as_model = __deserialize_book_data(book)
    assert books[0] == book_as_model


@pytest.mark.parametrize("store", redis_store_fixture)
def test_insert_single_nested(store):
    """
    Providing a single Model instance upserts also any nested model into redis
    """
    key = f"author_%&_{books[0].author.name}"
    author = store.redis_store.hgetall(name=key)
    assert author == {}

    Book.insert(books[0])

    author = store.redis_store.hgetall(name=key)
    author_as_model = Author(**Author.deserialize_partially(author))
    assert books[0].author == author_as_model


@pytest.mark.parametrize("store", redis_store_fixture)
def test_update_nested_list_of_models(store):
    data = [Library(name="Babel Library", address="In a book", books=books)]
    Library.insert(data)
    # the list of nested models is automatically inserted
    got = sorted(Book.select(), key=lambda x: x.title)
    expected = sorted(books, key=lambda x: x.title)
    assert expected == got

    got = sorted(Library.select(), key=lambda x: x.name)
    expected = sorted(data, key=lambda x: x.name)
    assert got == expected


@pytest.mark.parametrize("store", redis_store_fixture)
def test_update_optional_nested_list_of_models(store):
    data = [Library(name="Babel Library", address="In a book", lost=books)]
    Library.insert(data)
    # the list of nested models is automatically inserted
    got = sorted(Book.select(), key=lambda x: x.title)
    expected = sorted(books, key=lambda x: x.title)
    assert expected == got

    got = sorted(Library.select(), key=lambda x: x.name)
    expected = sorted(data, key=lambda x: x.name)
    assert got == expected


@pytest.mark.parametrize("store", redis_store_fixture)
def test_update_nested_tuple_of_models(store):
    jane = authors["jane"]
    new_stuff = (books[0], jane, books[1], 8)
    data = [Library(name="Babel Library", address="In a book", new=new_stuff)]
    Library.insert(data)
    # the tuple of nested models is automatically inserted
    got = sorted(Book.select(), key=lambda x: x.title)
    expected = sorted([books[0], books[1]], key=lambda x: x.title)
    assert expected == got

    got = sorted(Author.select(), key=lambda x: x.name)
    expected = sorted([books[0].author, jane], key=lambda x: x.name)
    assert expected == got

    got = sorted(Library.select(), key=lambda x: x.name)
    expected = sorted(data, key=lambda x: x.name)
    assert got == expected


@pytest.mark.parametrize("store", redis_store_fixture)
def test_update_optional_nested_tuple_of_models(store):
    popular_books = (books[0], books[2])
    data = [Library(name="Babel Library", address="In a book", popular=popular_books)]
    Library.insert(data)
    # the tuple of nested models is automatically inserted
    got = sorted(Book.select(), key=lambda x: x.title)
    expected = sorted(popular_books, key=lambda x: x.title)
    assert expected == got

    got = sorted(Library.select(), key=lambda x: x.name)
    expected = sorted(data, key=lambda x: x.name)
    assert got == expected


@pytest.mark.parametrize("store", redis_store_fixture)
def test_select_default(store):
    """Selecting without arguments returns all the book models"""
    Book.insert(books)
    response = Book.select()
    sorted_books = sorted(books, key=lambda x: x.title)
    sorted_response = sorted(response, key=lambda x: x.title)
    assert sorted_books == sorted_response


@pytest.mark.parametrize("store", redis_store_fixture)
def test_select_some_columns(store):
    """
    Selecting some columns returns a list of dictionaries of all books models with only those columns
    """
    Book.insert(books)
    books_dict = {book.title: book for book in books}
    columns = ["title", "author", "in_stock"]
    response = Book.select(columns=["title", "author", "in_stock"])
    response_dict = {book["title"]: book for book in response}

    for title, book in books_dict.items():
        book_in_response = response_dict[title]
        assert isinstance(book_in_response, dict)
        assert sorted(book_in_response.keys()) == sorted(columns)

        for column in columns:
            if column == "author":
                assert book_in_response[column] == getattr(book, column)
            else:
                assert f"{book_in_response[column]}" == f"{getattr(book, column)}"


@pytest.mark.parametrize("store", redis_store_fixture)
def test_select_some_ids(store):
    """
    Selecting some ids returns only those elements with the given ids
    """
    Book.insert(books)
    ids = [book.title for book in books[:2]]
    response = Book.select(ids=ids)
    assert response == books[:2]


@pytest.mark.parametrize("store", redis_store_fixture)
def test_update(store):
    """
    Updating an item of a given primary key updates it in redis
    """
    Book.insert(books)
    title = books[0].title
    new_in_stock = not books[0].in_stock
    new_author = Author(name="John Doe", active_years=(2000, 2009))
    book_key = f"book_%&_{title}"
    new_author_key = f"author_%&_{new_author.name}"
    old_book_data = store.redis_store.hgetall(name=book_key)
    old_book = __deserialize_book_data(old_book_data)
    assert old_book == books[0]
    assert old_book.author != new_author

    Book.update(_id=title, data={"author": new_author, "in_stock": new_in_stock})

    book_data = store.redis_store.hgetall(name=book_key)
    book = __deserialize_book_data(book_data)
    author_data = store.redis_store.hgetall(name=new_author_key)
    author = Author(**Author.deserialize_partially(author_data))
    assert book.author == new_author
    assert author == new_author
    assert book.title == old_book.title
    assert book.in_stock == new_in_stock
    assert book.published_on == old_book.published_on


@pytest.mark.parametrize("store", redis_store_fixture)
def test_update_nested_model(store):
    """
    Updating a nested model, without changing its primary key, also updates it its collection in redis
    """
    Book.insert(books)

    new_in_stock = not books[0].in_stock
    updated_author = Author(**books[0].author.dict())
    updated_author.active_years = (2020, 2045)
    book_key = f"book_%&_{books[0].title}"
    author_key = f"author_%&_{updated_author.name}"

    old_author_data = store.redis_store.hgetall(name=author_key)
    old_author = Author(**Author.deserialize_partially(old_author_data))
    old_book_data = store.redis_store.hgetall(name=book_key)
    old_book = __deserialize_book_data(old_book_data)
    assert old_book == books[0]
    assert old_author == books[0].author
    assert old_author != updated_author

    Book.update(
        _id=books[0].title, data={"author": updated_author, "in_stock": new_in_stock}
    )

    book_data = store.redis_store.hgetall(name=book_key)
    book = __deserialize_book_data(book_data)
    author_data = store.redis_store.hgetall(name=author_key)
    author = Author(**Author.deserialize_partially(author_data))
    assert book.author == updated_author
    assert author == updated_author
    assert book.title == old_book.title
    assert book.in_stock == new_in_stock
    assert book.published_on == old_book.published_on


@pytest.mark.parametrize("store", redis_store_fixture)
def test_delete_multiple(store):
    """
    Providing a list of ids to the delete function will remove the items from redis,
    but leave the nested models intact
    """
    Book.insert(books)
    books_to_delete = books[:2]
    books_left_in_db = books[2:]

    ids_to_delete = [book.title for book in books_to_delete]
    ids_to_leave_intact = [book.title for book in books_left_in_db]

    keys_to_delete = [f"book_%&_{_id}" for _id in ids_to_delete]
    book_keys_to_leave_intact = [f"book_%&_{_id}" for _id in ids_to_leave_intact]
    author_keys_to_leave_intact = [
        f"author_%&_{author.name}" for author in authors.values()
    ]

    Book.delete(ids=ids_to_delete)

    for key in keys_to_delete:
        deleted_book_in_redis = store.redis_store.hgetall(name=key)
        assert deleted_book_in_redis == {}

    with store.redis_store.pipeline() as pipeline:
        for key in book_keys_to_leave_intact:
            pipeline.hgetall(name=key)
        books_in_redis = pipeline.execute()
        books_in_redis_as_models = [
            __deserialize_book_data(book) for book in books_in_redis
        ]
        assert books_left_in_db == books_in_redis_as_models

        for key in author_keys_to_leave_intact:
            pipeline.hgetall(name=key)
        authors_in_redis = pipeline.execute()
        authors_in_redis_as_models = sorted(
            [
                Author(**Author.deserialize_partially(author))
                for author in authors_in_redis
            ],
            key=lambda x: x.name,
        )
        expected = sorted(authors.values(), key=lambda x: x.name)
        assert expected == authors_in_redis_as_models


def __deserialize_book_data(raw_book_data: Dict[str, Any]) -> Book:
    """Deserializes the raw book data returning a book instance"""
    author_id = raw_book_data.pop(f"{NESTED_MODEL_PREFIX}author")
    author_id = strip_leading(author_id, "author_%&_")

    data = Book.deserialize_partially(raw_book_data)

    data["author"] = Author.select(ids=[author_id])[0]
    return Book(**data)
